from typing import MutableSequence, Optional
from itertools import chain

import attr

from vkit.label.type import VPoint, VPolygon, VPointList


@attr.define
class VImageGrid:
    points_2d: MutableSequence[MutableSequence[VPoint]]

    _cache_image_height: Optional[int] = None
    _cache_image_width: Optional[int] = None

    @property
    def num_rows(self):
        return len(self.points_2d)

    @property
    def num_cols(self):
        return len(self.points_2d[0])

    @property
    def flatten_points(self):
        return VPointList(chain.from_iterable(self.points_2d))

    @property
    def image_height(self):
        if self._cache_image_height is None:
            assert min(point.y for point in self.flatten_points) == 0
            self._cache_image_height = max(point.y for point in self.flatten_points) + 1
        return self._cache_image_height

    @property
    def image_width(self):
        if self._cache_image_width is None:
            assert min(point.x for point in self.flatten_points) == 0
            self._cache_image_width = max(point.x for point in self.flatten_points) + 1
        return self._cache_image_width

    @property
    def shape(self):
        return self.num_rows, self.num_cols

    def compatible_with(self, other):
        return self.shape == other.shape

    def generate_polygons(self):
        for row in range(1, self.num_rows):
            for col in range(1, self.num_cols):
                yield VPolygon(
                    points=VPointList([
                        # Clockwise.
                        self.points_2d[row - 1][col - 1],
                        self.points_2d[row - 1][col],
                        self.points_2d[row][col],
                        self.points_2d[row][col - 1],
                    ]),
                )

    def zip_polygons(self, other):
        assert self.compatible_with(other)
        for self_polygon, other_polygon in zip(self.generate_polygons(), other.generate_polygons()):
            yield self_polygon, other_polygon

    def generate_border_polygon(self):
        # Clockwise.
        points = VPointList()

        for point in self.points_2d[0]:
            points.append(point)
        for row in range(1, self.num_rows):
            points.append(self.points_2d[row][-1])
        for col in reversed(range(self.num_cols - 1)):
            points.append(self.points_2d[-1][col])
        for row in reversed(range(1, self.num_rows - 1)):
            points.append(self.points_2d[row][0])

        return VPolygon(points=points)
